import inspect
from ds_capability.components.commons import Commons
from ds_core.intent.abstract_intent import AbstractIntentModel

from ds_capability import FeatureBuild, FeatureTransform, FeatureSelect, FeatureEngineer, AutoML
from ds_capability.managers.controller_property_manager import ControllerPropertyManager


class ControllerIntentModel(AbstractIntentModel):

    """This component provides a set of actions that focuses on the Controller. The Controller is a unique component
    that independently orchestrates the components registered to it. It executes the components Domain Contract and
    not its code. The Controller orchestrates how those components should run with the components being independent
    in their actions and therefore a separation of concerns."""

    def __init__(self, property_manager: ControllerPropertyManager, default_save_intent: bool=None,
                 default_intent_level: [str, int, float]=None, order_next_available: bool=None,
                 default_replace_intent: bool=None):
        """initialisation of the Intent class.

        :param property_manager: the property manager class that references the intent contract.
        :param default_save_intent: (optional) The default action for saving intent in the property manager
        :param default_intent_level: (optional) the default level intent should be saved at
        :param order_next_available: (optional) if the default behaviour for the order should be next available order
        :param default_replace_intent: (optional) the default replace existing intent behaviour
        """
        default_save_intent = default_save_intent if isinstance(default_save_intent, bool) else True
        default_replace_intent = default_replace_intent if isinstance(default_replace_intent, bool) else True
        default_intent_level = default_intent_level if isinstance(default_intent_level, (str, int, float)) else 'base'
        default_intent_order = -1 if isinstance(order_next_available, bool) and order_next_available else 0
        intent_param_exclude = ['canonical']
        intent_type_additions = []
        super().__init__(property_manager=property_manager, default_save_intent=default_save_intent,
                         intent_param_exclude=intent_param_exclude, default_intent_level=default_intent_level,
                         default_intent_order=default_intent_order, default_replace_intent=default_replace_intent,
                         intent_type_additions=intent_type_additions)

    def run_intent_pipeline(self, run_level: str, source: str=None, persist: [str, list]=None,
                            controller_repo: str=None, **kwargs):
        """ Collectively runs all parameterised intent taken from the property manager against the code base as
        defined by the intent_contract.

        It is expected that all intent methods have the 'canonical' as the first parameter of the method signature
        and will contain 'save_intent' as parameters.

        :param run_level:
        :param persist:
        :param source:
        :param controller_repo: (optional) the controller repo to use if no uri_pm_repo is within the intent parameters
        :param kwargs: additional kwargs to add to the parameterised intent, these will replace any that already exist
        :return: Canonical with parameterised intent applied
        """
        # get the list of levels to run
        if not self._pm.has_intent(run_level):
            raise ValueError(f"The intent level '{run_level}' could not be found in the "
                             f"property manager '{self._pm.manager_name()}' for task '{self._pm.task_name}'")
        shape = None
        level_key = self._pm.join(self._pm.KEY.intent_key, run_level)
        for order in sorted(self._pm.get(level_key, {})):
            for method, params in self._pm.get(self._pm.join(level_key, order), {}).items():
                if method in self.__dir__():
                    # failsafe in case kwargs was stored as the reference
                    params.update(params.pop('kwargs', {}))
                    # add method kwargs to the params
                    if isinstance(kwargs, dict):
                        params.update(kwargs)
                    # remove the creator param
                    _ = params.pop('intent_creator', 'Unknown')
                    # add excluded params and set to False
                    params.update({'save_intent': False})
                    # add the controller_repo if given
                    if isinstance(controller_repo, str) and 'uri_pm_repo' not in params.keys():
                        params.update({'uri_pm_repo': controller_repo})
                    shape = eval(f"self.{method}(source=source, persist=persist, **{params})", globals(), locals())
        return shape

    def feature_build(self, task_name: str, source: str=None, persist: [str, list]=None, columns: [str, list]=None,
                      seed: int=None, save_intent: bool=None, intent_order: int=None, intent_level: [int, str]=None,
                      replace_intent: bool=None, remove_duplicates: bool=None, **kwargs):
        """ register a Transition component task pipeline

        :param persist:
        :param source:
        :param task_name: the task_name reference for this component
        :param columns: (optional) a single or list of intent_level to run, if list, run in order given
        :param seed: (optional) a seed for the run
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                        If None: default's to -1
                        if -1: added to a level above any current instance of the intent section, level 0 if not found
                        if int: added to the level specified, overwriting any that already exist
        :param replace_intent: (optional) if the intent method exists at the level, or default level
                        True - replaces the current intent method with the new
                        False - leaves it untouched, disregarding the new intent
        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
       """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # create the event book
        fb: FeatureBuild = eval(f"FeatureBuild.from_env(task_name=task_name, default_save=False, "
                                f"has_contract=True, **{kwargs})", globals(), locals())
        if source and fb.pm.has_connector(source):
            canonical = fb.load_canonical(source)
        elif fb.pm.has_connector(fb.CONNECTOR_SOURCE):
            canonical = fb.load_source_canonical()
        else:
            canonical = None
        canonical = fb.intent_model.run_intent_pipeline(canonical=canonical, intent_levels=intent_level, seed=seed)
        if persist:
            for out in Commons.list_formatter(persist):
                if fb.pm.has_connector(out):
                    fb.save_canonical(connector_name=out, canonical=canonical)
        else:
            fb.save_persist_canonical(canonical=canonical)
        return canonical.shape

    def feature_engineer(self, task_name: str, source: str=None, persist: [str, list]=None, columns: [str, list]=None,
                         seed: int=None, save_intent: bool=None, intent_order: int=None, intent_level: [int, str]=None,
                         replace_intent: bool=None, remove_duplicates: bool=None, **kwargs):
        """ register a Transition component task pipeline

        :param persist:
        :param source:
        :param task_name: the task_name reference for this component
        :param columns: (optional) a single or list of intent_level to run, if list, run in order given
        :param seed: (optional) a seed for the run
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                        If None: default's to -1
                        if -1: added to a level above any current instance of the intent section, level 0 if not found
                        if int: added to the level specified, overwriting any that already exist
        :param replace_intent: (optional) if the intent method exists at the level, or default level
                        True - replaces the current intent method with the new
                        False - leaves it untouched, disregarding the new intent
        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
       """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # create the event book
        fe: FeatureEngineer = eval(f"FeatureEngineer.from_env(task_name=task_name, default_save=False, "
                                   f"has_contract=True, **{kwargs})", globals(), locals())
        if source and fe.pm.has_connector(source):
            canonical = fe.load_canonical(source)
        elif fe.pm.has_connector(fe.CONNECTOR_SOURCE):
            canonical = fe.load_source_canonical()
        else:
            canonical = None
        canonical = fe.intent_model.run_intent_pipeline(canonical=canonical, intent_levels=intent_level, seed=seed)
        if persist:
            for out in Commons.list_formatter(persist):
                if fe.pm.has_connector(out):
                    fe.save_canonical(connector_name=out, canonical=canonical)
        else:
            fe.save_persist_canonical(canonical=canonical)
        return canonical.shape

    def feature_transform(self, task_name: str, source: str=None, persist: [str, list]=None, columns: [str, list]=None,
                          seed: int=None, save_intent: bool=None, intent_order: int=None, intent_level: [int, str]=None,
                          replace_intent: bool=None, remove_duplicates: bool=None, **kwargs):
        """ register a Transition component task pipeline

        :param persist:
        :param source:
        :param task_name: the task_name reference for this component
        :param columns: (optional) a single or list of intent_level to run, if list, run in order given
        :param seed: (optional) a seed for the run
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                        If None: default's to -1
                        if -1: added to a level above any current instance of the intent section, level 0 if not found
                        if int: added to the level specified, overwriting any that already exist
        :param replace_intent: (optional) if the intent method exists at the level, or default level
                        True - replaces the current intent method with the new
                        False - leaves it untouched, disregarding the new intent
        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
       """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # create the event book
        ft: FeatureTransform = eval(f"FeatureTransform.from_env(task_name=task_name, default_save=False, "
                                    f"has_contract=True, **{kwargs})", globals(), locals())
        if source and ft.pm.has_connector(source):
            canonical = ft.load_canonical(source)
        elif ft.pm.has_connector(ft.CONNECTOR_SOURCE):
            canonical = ft.load_source_canonical()
        else:
            canonical = None
        canonical = ft.intent_model.run_intent_pipeline(canonical=canonical, intent_levels=intent_level, seed=seed)
        if persist:
            for out in Commons.list_formatter(persist):
                if ft.pm.has_connector(out):
                    ft.save_canonical(connector_name=out, canonical=canonical)
        else:
            ft.save_persist_canonical(canonical=canonical)
        return canonical.shape

    def feature_select(self, task_name: str, source: str=None, persist: [str, list]=None, columns: [str, list]=None,
                       seed: int=None, save_intent: bool=None, intent_order: int=None, intent_level: [int, str]=None,
                       replace_intent: bool=None, remove_duplicates: bool=None, **kwargs):
        """ register a select component task pipeline

        :param persist:
        :param source:
        :param task_name: the task_name reference for this component
        :param columns: (optional) a single or list of intent_level to run, if list, run in order given
        :param seed: (optional) a seed for the run
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                        If None: default's to -1
                        if -1: added to a level above any current instance of the intent section, level 0 if not found
                        if int: added to the level specified, overwriting any that already exist
        :param replace_intent: (optional) if the intent method exists at the level, or default level
                        True - replaces the current intent method with the new
                        False - leaves it untouched, disregarding the new intent
        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
       """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # create the event book
        fs: FeatureSelect = eval(f"FeatureSelect.from_env(task_name=task_name, default_save=False, "
                                 f"has_contract=True, **{kwargs})", globals(), locals())
        if source and fs.pm.has_connector(source):
            canonical = fs.load_canonical(source)
        elif fs.pm.has_connector(fs.CONNECTOR_SOURCE):
            canonical = fs.load_source_canonical()
        else:
            canonical = None
        canonical = fs.intent_model.run_intent_pipeline(canonical=canonical, intent_levels=intent_level, seed=seed)
        if persist:
            for out in Commons.list_formatter(persist):
                if fs.pm.has_connector(out):
                    fs.save_canonical(connector_name=out, canonical=canonical)
        else:
            fs.save_persist_canonical(canonical=canonical)
        return canonical.shape

    def automl(self, task_name: str, source: str=None, persist: [str, list]=None, columns: [str, list]=None,
               seed: int=None, save_intent: bool=None, intent_order: int=None, intent_level: [int, str]=None,
               replace_intent: bool=None, remove_duplicates: bool=None, **kwargs):
        """ register an automl component task pipeline

        :param persist:
        :param source:
        :param task_name: the task_name reference for this component
        :param columns: (optional) a single or list of intent_level to run, if list, run in order given
        :param seed: (optional) a seed for the run
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                        If None: default's to -1
                        if -1: added to a level above any current instance of the intent section, level 0 if not found
                        if int: added to the level specified, overwriting any that already exist
        :param replace_intent: (optional) if the intent method exists at the level, or default level
                        True - replaces the current intent method with the new
                        False - leaves it untouched, disregarding the new intent
        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
       """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # create the event book
        aml: AutoML = eval(f"AutoML.from_env(task_name=task_name, default_save=False, "
                                 f"has_contract=True, **{kwargs})", globals(), locals())
        if source and aml.pm.has_connector(source):
            canonical = aml.load_canonical(source)
        elif aml.pm.has_connector(aml.CONNECTOR_SOURCE):
            canonical = aml.load_source_canonical()
        else:
            canonical = None
        canonical = aml.intent_model.run_intent_pipeline(canonical=canonical, intent_levels=intent_level, seed=seed)
        if persist:
            for out in Commons.list_formatter(persist):
                if aml.pm.has_connector(out):
                    aml.save_canonical(connector_name=out, canonical=canonical)
        else:
            aml.save_persist_canonical(canonical=canonical)
        return canonical.shape

    def _set_intend_signature(self, intent_params: dict, intent_level: [int, str]=None, intent_order: int=None,
                              replace_intent: bool=None, remove_duplicates: bool=None, save_intent: bool=None):
        """ sets the intent section in the configuration file. Note: by default any identical intent, e.g.
        intent with the same intent (name) and the same parameter values, are removed from any level.

        :param intent_params: a dictionary type set of configuration representing a intent section contract
        :param intent_level: (optional) ta level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                        If None: default's to -1
                        if -1: added to a level above any current instance of the intent section, level 0 if not found
                        if int: added to the level specified, overwriting any that already exist
        :param replace_intent: (optional) if the intent method exists at the level, or default level
                        True - replaces the current intent method with the new
                        False - leaves it untouched, disregarding the new intent
        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :param save_intent (optional) if the intent contract should be saved to the property manager
        """
        intent_level = intent_level if isinstance(intent_level, (str, int)) else self._pm.DEFAULT_INTENT_LEVEL
        if save_intent or (not isinstance(save_intent, bool) and self._default_save_intent):
            if not isinstance(intent_order, int) or intent_order == -1:
                if self._pm.get_intent():
                    intent_order = 0
                    while True:
                        if not self._pm.is_key(self._pm.join(self._pm.KEY.intent_key, intent_level, intent_order)):
                            break
                        intent_order += 1
        super()._set_intend_signature(intent_params=intent_params, intent_level=intent_level, intent_order=intent_order,
                                      replace_intent=replace_intent, remove_duplicates=remove_duplicates,
                                      save_intent=save_intent)
        return



import unittest
import os
from pathlib import Path
import shutil
import ast
from datetime import datetime
from pprint import pprint

import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.parquet as pq
from ds_capability.components.commons import Commons
from ds_core.properties.property_manager import PropertyManager
from ds_capability import Controller, FeatureBuild
from ds_capability.intent.controller_intent import ControllerIntentModel

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class FeatureBuilderTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_controller_runbook(self):
        # tbl = FeatureBuild.from_memory().tools.get_synthetic_data_types(size=10)
        controller: Controller = Controller.from_env('control', has_contract=False)
        register: ControllerIntentModel = controller.register
        fb1 = set_service_one()
        fb2 = set_service_two(fb1)
        # build controller
        fb_tbl = register.feature_build('task1', intent_level='fb1')
        fb_tbl = register.feature_build('task2', intent_level='fb2')
        controller.add_run_book_level(book_name='test_item', run_level='fb1', source='event1', persist='event1')
        controller.add_run_book_level(book_name='test_item', run_level='fb2', source='event1', persist='event2')
        controller.add_run_book_level(book_name='test_item', run_level='fb2', source='event1', persist=['event2', 'event3'])
        controller.add_run_book(run_levels=['fb1', 'fb2'], book_name='test_book')
        # pprint(pm_view('controller', 'control', 'run_book'))
        # pprint(pm_view('controller', 'control', 'intent'))
        controller.run_controller(run_book=['test_item', 'test_book'], run_cycle_report='tmp_report.parquet')
        result = controller.load_canonical('run_cycle_report')
        print(Commons.table_report(result).to_string())


    def test_controller(self):
        # tbl = FeatureBuild.from_memory().tools.get_synthetic_data_types(size=10)
        controller: Controller = Controller.from_env('control', has_contract=False)
        register: ControllerIntentModel = controller.register
        fb1 = set_service_one()
        fb2 = set_service_two(fb1)
        # build controller
        register.feature_build('task1')
        register.feature_build('task2', intent_level='pre-process')
        # pprint(pm_view('controller', 'control', 'intent'))
        controller.run_controller(run_cycle_report='tmp_report.parquet')
        result = controller.load_canonical('run_cycle_report')
        print(Commons.table_report(result).to_string())


    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")


def set_service_one():
    fb = FeatureBuild.from_env('task1', has_contract=False)
    fb.set_persist_uri("event://task1_outcome")
    tbl = fb.tools.get_synthetic_data_types(size=10)
    _ = fb.tools.get_noise(size=10, num_columns=2, canonical=tbl)
    # pprint(pm_view('feature_build', 'task1'))
    fb.run_component_pipeline()
    return fb

def set_service_two(other: FeatureBuild):
    fb = FeatureBuild.from_env('task2', has_contract=False)
    fb.set_source_uri(other.get_persist_uri())
    fb.set_persist_uri("event://task2_outcome")
    source_tbl = fb.load_source_canonical()
    tbl = fb.tools.correlate_number(canonical=source_tbl, header='num', jitter=2, to_header='jitter')
    _ = fb.tools.correlate_column_join(tbl, header='PI', others='int', to_header='id')
    # pprint(pm_view('feature_build', 'task2'))
    fb.run_component_pipeline()
    return fb


def pm_view(capability: str, task: str, section: str=None):
    uri = os.path.join(os.environ['HADRON_PM_PATH'], f"hadron_pm_{capability}_{task}.parquet")
    tbl = pq.read_table(uri)
    tbl = tbl.column(0).combine_chunks()
    result = ast.literal_eval(tbl.to_pylist()[0]).get(capability, {}).get(task, {})
    return result.get(section, {}) if isinstance(section, str) and section in result.keys() else result


if __name__ == '__main__':
    unittest.main()



import math
import random
import pyarrow as pa
import pyarrow.compute as pc
import pandas as pd
import numpy as np
from matplotlib.colors import LogNorm

from ds_capability.components.commons import Commons
from scipy import stats
from sklearn.model_selection import train_test_split
from matplotlib import pyplot as plt
import seaborn as sns


class Visualisation(object):
    """ a set of data components methods to Visualise pandas.Dataframe"""

    @staticmethod
    def show_chi_square(canonical: pa.Table, target: str, capped_at: int=None, seed: int=None, width: float=None, height: float=None):
        """ Chi-square is one of the most widely used supervised feature selection methods. It selects each feature
         independently in accordance with their scores against a target or label then ranks them by their importance.
         This score should be used to evaluate categorical variables in a classification task.

        :param canonical: The canonical to apply
        :param target: the str header that constitute a binary target.
        :param capped_at: (optional) a cap on the size of elements (columns x rows) to process. default at 5,000,000
        :param seed: (optional) a seed value for the test train dataset
        :param width: (optional) the figure size width
        :param height: (optional) the figure size height
        """
        if target not in canonical.column_names:
            raise ValueError(f"The target '{target}' can't be found in the canonical")
        if pc.count(pc.unique(canonical.column(target))).as_py() != 2:
            raise ValueError(f"The target '{target}' must only be two unique values")
        width = width if isinstance(width, float) else 6
        height = height if isinstance(height, float) else 4
        cap = capped_at if isinstance(capped_at, int) else 1_000_000
        if canonical.num_rows*canonical.num_columns > cap > 0:
            sample = random.sample(range(canonical.num_rows), k=int(cap/canonical.num_columns))
            canonical = canonical.take(sample)
        control = canonical.to_pandas()
        # separate train and test sets
        X_train, X_test, y_train, y_test = train_test_split(control.drop(target, axis=1), control[target],
                                                            test_size=0.3, random_state=seed)
        chi_ls = []
        for feature in X_train.columns:
            # create contingency table
            c = pd.crosstab(y_train, X_train[feature])
            # chi_test
            p_value = stats.chi2_contingency(c)[1]
            chi_ls.append(p_value)
        plt.figure(figsize=(width, height))
        pd.Series(chi_ls, index=X_train.columns).sort_values(ascending=True).plot.bar(rot=45)
        plt.ylabel('p value')
        plt.title('Feature importance based on chi-square test', fontdict={'size': 20})
        plt.tight_layout()
        plt.show()
        plt.clf()

    @staticmethod
    def show_missing(canonical: pa.Table, headers: [str, list]=None, d_types: [str, list]=None,
                     regex: [str, list]=None, drop: bool=None, capped_at: int=None, width: float=None, height: float=None):
        """ A heatmap of missing data. Each column shows missing values and where in the column the missing data is.

        :param canonical: a canonical to apply
        :param headers: (optional) a filter on headers from the canonical
        :param d_types: (optional) a filter on data type for the canonical. example [pa.int64(), pa.string()]
        :param regex: (optional) a regular expression to search the headers.
        :param drop: (optional) to drop or not drop the resulting headers.
        :param capped_at: (optional) a cap on the size of elements (columns x rows) to process. default at 5,000,000
        :param width: (optional) the figure size width
        :param height: (optional) the figure size height
        """
        width = width if isinstance(width, float) else 6
        height = height if isinstance(height, float) else 5
        cap = capped_at if isinstance(capped_at, int) else 1_000_000
        if canonical.num_rows*canonical.num_columns > cap > 0:
            sample = random.sample(range(canonical.num_rows), k=int(cap/canonical.num_columns))
            canonical = canonical.take(sample)
        canonical = Commons.filter_columns(canonical, headers=headers, d_types=d_types, regex=regex, drop=drop)
        control = canonical.to_pandas()
        sns.set(rc={'figure.figsize': (width, height)})
        sns.heatmap(control.isnull(), yticklabels=False, cbar=False, cmap='viridis')
        plt.title('missing data', fontdict={'size': 20})
        plt.tight_layout()
        plt.show()
        plt.clf()

    @staticmethod
    def show_correlated(canonical: pa.Table, headers: [str, list]=None, d_types: [str, list]=None,
                        regex: [str, list]=None, drop: bool=None, capped_at: int=None, width: float=None, height: float=None):
        """ shows correlation as a grid of values for each column pair where correlation is represented by a value
        moving towards 100. This only applies to int and float.

        :param canonical: a canonical to apply
        :param headers: (optional) a filter on headers from the canonical
        :param d_types: (optional) a filter on data type for the canonical. example [pa.int64(), pa.string()]
        :param regex: (optional) a regular expression to search the headers.
        :param drop: (optional) to drop or not drop the resulting headers.
        :param capped_at: (optional) a cap on the size of elements (columns x rows) to process. default at 5,000,000
        :param width: (optional) the figure size width
        :param height: (optional) the figure size height
        :return: pa.Table
        """
        width = width if isinstance(width, float) else 6
        height = height if isinstance(height, float) else 5
        cap = capped_at if isinstance(capped_at, int) else 1_000_000
        if canonical.num_rows*canonical.num_columns > cap > 0:
            sample = random.sample(range(canonical.num_rows), k=int(cap/canonical.num_columns))
            canonical = canonical.take(sample)
        canonical = Commons.filter_columns(canonical, headers=headers, d_types=d_types, regex=regex, drop=drop)
        canonical = Commons.filter_columns(canonical, d_types=[pa.int64(),pa.int32(),pa.int16(),pa.int8(),
                                                               pa.float64(),pa.float32(),pa.float16()])
        control = canonical.to_pandas()
        sns.set(rc={'figure.figsize': (width, height)})
        sns.heatmap(control.corr(), annot=True, cmap='BuGn', robust=True)
        plt.title('correlated data', fontdict={'size': 20})
        plt.tight_layout()
        plt.show()
        plt.clf()

    @staticmethod
    def show_distributions(canonical: pa.Table, headers: [str, list]=None, d_types: [str, list]=None,
                        regex: [str, list]=None, drop: bool=None, capped_at: int=None, width: float=None, height: float=None):
        """ Shows three key distributions for a target sample array."""
        width = width if isinstance(width, float) else 16
        height = height if isinstance(height, float) else 4
        cap = capped_at if isinstance(capped_at, int) else 1_000_000
        if canonical.num_rows*canonical.num_columns > cap > 0:
            sample = random.sample(range(canonical.num_rows), k=int(cap/canonical.num_columns))
            canonical = canonical.take(sample)
        canonical = Commons.filter_columns(canonical, headers=headers, d_types=d_types, regex=regex, drop=drop)
        canonical = Commons.filter_columns(canonical, d_types=[pa.int64(),pa.int32(),pa.int16(),pa.int8(),
                                                               pa.float64(),pa.float32(),pa.float16()])
        control = canonical.to_pandas()
        for target in canonical.column_names:
        # Define figure size.
            _ = plt.figure(figsize=(width, height))
            _ = plt.suptitle('Show Distribution', fontdict={'size': 20})
            # histogram
            plt.subplot(1, 3, 1)
            sns.histplot(control[target], bins=30)
            plt.title('Histogram')
            # Q-Q plot
            plt.subplot(1, 3, 2)
            stats.probplot(control[target], dist="norm", plot=plt)
            plt.ylabel('RM quantiles')
            # boxplot
            plt.subplot(1, 3, 3)
            sns.boxplot(y=control[target])
            plt.title('Boxplot')
        plt.tight_layout()
        plt.show()
        plt.clf()

    @staticmethod
    def show_category_frequency(canonical: pa.Table, target_dt, headers: [str, list]=None, d_types: [str, list]=None,
                                regex: [str, list]=None, drop: bool=None, category_limit: int=None, capped_at: int=None, log_scale=False,
                                subplot_h=2, subplot_w=15, param_scale=8, rotation=360, hspace=0.35):
        """ creates the frequencies (colors of heatmap) of the elements (y axis) of the categorical columns
        over time (x axis)

        :param canonical:
        :param target_dt: the target date
        :param headers: (optional) a filter on headers from the canonical
        :param d_types: (optional) a filter on data type for the canonical. example [pa.int64(), pa.string()]
        :param regex: (optional) a regular expression to search the headers.
        :param drop: (optional) to drop or not drop the resulting headers.
        :param category_limit: (optional) the number of unique values that make up a category. Default 10
        :param capped_at: (optional) a cap on the size of elements (columns x rows) to process. default at 5,000,000
        :param log_scale: (optional)
        :param subplot_h: (optional)
        :param subplot_w: (optional)
        :param param_scale: (optional)
        :param rotation: (optional)
        :param hspace: (optional)
        """
        cap = capped_at if isinstance(capped_at, int) else 1_000_000
        category_limit = category_limit if isinstance(category_limit, str) else 10
        log_scale = log_scale if isinstance(log_scale, bool) else False
        subplot_h = subplot_h if isinstance(subplot_h, int) else 2
        subplot_w = subplot_w if isinstance(subplot_w, int) else 15
        param_scale = param_scale if isinstance(param_scale, int) else 0
        rotation = rotation if isinstance(rotation, int) else 360
        hspace = hspace if isinstance(hspace, float) else 0.35
        if canonical.num_rows*canonical.num_columns > cap > 0:
            sample = random.sample(range(canonical.num_rows), k=int(cap/canonical.num_columns))
            canonical = canonical.take(sample)
        canonical = Commons.filter_columns(canonical, headers=headers, d_types=d_types, regex=regex, drop=drop)
        col_names = Commons.filter_headers(canonical, d_types=[pa.string()])
        for n in tuple(col_names):
            c = canonical.column(n).combine_chunks()
            if pc.count(pc.unique(c)).as_py() > category_limit:
                col_names.remove(n)
        if target_dt in col_names:
            col_names.remove(target_dt)
        df = canonical.to_pandas()
        df[target_dt] = pd.to_datetime(df[target_dt])
        dates = pd.date_range(start=df[target_dt].min(), end=df[target_dt].max())
        n_categories = len(col_names)
        cbar_kws = {'orientation': 'horizontal', 'shrink': 0.5}
        n_subplot_rows = np.ceil(df[col_names].nunique(dropna=True).divide(param_scale))
        n_subplot_rows.iloc[-1] += 1
        n_rows = int(n_subplot_rows.sum())
        grid_weights = {'height_ratios': n_subplot_rows.values}
        cmap = 'rocket_r'
        # cmap = sns.cm.rocket_r
        fig, axes = plt.subplots(n_categories, 1, gridspec_kw=grid_weights, sharex='col',
                                 figsize=(subplot_w, n_rows * subplot_h))
        if n_categories == 1:
            axes = [axes]
        for ii in range(n_categories):
            cc = col_names[ii]
            df_single_cat = df[[target_dt, cc]]
            df_single_cat = df_single_cat.loc[df_single_cat[target_dt].notnull(),]
            df_single_cat['Index'] = df_single_cat[target_dt].dt.date
            df_pivot = df_single_cat.pivot_table(index='Index', columns=cc, aggfunc=len, dropna=True)
            df_pivot.index = pd.to_datetime(df_pivot.index)
            toplot = df_pivot.reindex(dates.date).T

            v_min = toplot.min().min()
            v_max = toplot.max().max()
            toplot.reset_index(level=0, drop=True, inplace=True)
            if log_scale:
                cbar_ticks = [math.pow(10, i) for i in range(int(math.floor(math.log10(v_min))),
                                                             int(1 + math.ceil(math.log10(v_max))))]
                log_norm = LogNorm(vmin=v_min, vmax=v_max)
            else:
                cbar_ticks = list(range(int(v_min), int(v_max + 1)))
                if len(cbar_ticks) > 5:
                    v_step = int(math.ceil((v_max - v_min) / 4))
                    cbar_ticks = list(range(int(v_min), int(v_max + 1), v_step))
                log_norm = None
            cbar_kws['ticks'] = cbar_ticks
            if ii < (n_categories - 1):
                cbar_kws['pad'] = 0.05
            else:
                cbar_kws['pad'] = 0.25
            sns.heatmap(toplot, cmap=cmap, ax=axes[ii], norm=log_norm, cbar_kws=cbar_kws, yticklabels=True)
            axes[ii].set_ylabel('')
            axes[ii].set_xlabel('')
            axes[ii].set_title(cc)
            axes[ii].set_yticklabels(axes[ii].get_yticklabels(), rotation=rotation)
            for _, spine in axes[ii].spines.items():
                spine.set_visible(True)
        axes[-1].set_xlabel(target_dt)
        plt.subplots_adjust(bottom=0.05, hspace=hspace)
        plt.tight_layout()
        plt.show()
        plt.clf()

    @staticmethod
    def show_category_appearance(canonical: pa.Table, target_dt: str, headers: [str, list]=None,
                                 d_types: [str, list]=None, regex: [str, list]=None, drop: bool=None,
                                 category_limit: int=None, capped_at: int=None, subplot_h: int=None,
                                 subplot_w: int=None, rotation: int=None):
        """ creates the proportion (as percentages) (colors of heatmap) of the appearing elements (y axis)
        of the categorical columns over time (x axis)

        :param canonical:
        :param target_dt:
        :param headers: (optional) a filter on headers from the canonical
        :param d_types: (optional) a filter on data type for the canonical. example [pa.int64(), pa.string()]
        :param regex: (optional) a regular expression to search the headers.
        :param drop: (optional) to drop or not drop the resulting headers.
        :param category_limit: (optional) the number of unique values that make up a category. Default 10
        :param capped_at: (optional) a cap on the size of elements (columns x rows) to process. default at 5,000,000
        :param subplot_h: (optional)
        :param subplot_w: (optional)
        :param rotation: (optional)
        """
        category_limit = category_limit if isinstance(category_limit, str) else 10
        cap = capped_at if isinstance(capped_at, int) else 1_000_000
        subplot_h = subplot_h if isinstance(subplot_h, int) else 6
        subplot_w = subplot_w if isinstance(subplot_w, int) else 10
        rotation = rotation if isinstance(rotation, int) else 360
        if canonical.num_rows*canonical.num_columns > cap > 0:
            sample = random.sample(range(canonical.num_rows), k=int(cap/canonical.num_columns))
            canonical = canonical.take(sample)
        canonical = Commons.filter_columns(canonical, headers=headers, d_types=d_types, regex=regex, drop=drop)
        col_names = Commons.filter_headers(canonical, d_types=[pa.string()])
        for n in tuple(col_names):
            c = canonical.column(n).combine_chunks()
            if pc.count(pc.unique(c)).as_py() > category_limit:
                col_names.remove(n)
        if target_dt in col_names:
            col_names.remove(target_dt)
        df = canonical.to_pandas()
        df[target_dt] = pd.to_datetime(df[target_dt])
        dates = pd.date_range(start=df[target_dt].min(), end=df[target_dt].max())
        cmap = 'rocket_r'
        # cmap = sns.cm.rocket_r
        df0 = df[col_names + [target_dt]]
        df0['Index'] = df0.loc[:,target_dt].dt.date
        df_unique = df0[col_names].nunique(dropna=True)
        df_agg = df0.groupby('Index').nunique(dropna=True)
        # df_agg = df_agg.drop('Index', axis='columns')
        df_frac = df_agg[col_names].divide(df_unique, axis=1)
        df_frac.index = pd.to_datetime(df_frac.index)
        toplot = df_frac.reindex(dates.date).T
        new_labels = df_unique.index.values + '\n(' + pd.Series(df_unique.values).apply(str) + ')'
        fig = plt.figure(figsize=(subplot_w, subplot_h))
        ax = sns.heatmap(toplot, cmap=cmap, vmin=0, vmax=1, cbar_kws={'shrink': 0.75})
        ax.set_yticklabels(new_labels, rotation=rotation)
        ax.set_ylabel('')
        ax.set_xlabel(target_dt)
        cbar = ax.collections[0].colorbar
        cbar.set_ticks([0.0, 0.25, 0.5, 0.75, 1.0])
        cbar.set_ticklabels(['0%', '25%', '50%', '75%', '100%'])
        plt.tight_layout()
        plt.show()
        plt.clf()

    @staticmethod
    def show_numeric_density(canonical: pa.Table, headers: [str, list]=None, d_types: [str, list]=None,
                             regex: [str, list]=None, drop: bool=None, capped_at: int=None, width: float=None, height: float=None):
        """"""
        width = width if isinstance(width, float) else 16
        height = height if isinstance(height, float) else 4
        cap = capped_at if isinstance(capped_at, int) else 1_000_000
        if canonical.num_rows*canonical.num_columns > cap > 0:
            sample = random.sample(range(canonical.num_rows), k=int(cap/canonical.num_columns))
            canonical = canonical.take(sample)
        canonical = Commons.filter_columns(canonical, headers=headers, d_types=d_types, regex=regex, drop=drop)
        num_cols = Commons.filter_headers(canonical, d_types=[pa.int64(),pa.int32(),pa.int16(),pa.int8(),
                                                               pa.float64(),pa.float32(),pa.float16()])
        control = canonical.to_pandas()
        depth = int(round(len(num_cols) / 2, 0) + len(num_cols) % 2)
        _figsize = (width, height * depth)
        fig = plt.figure(figsize=_figsize)
        right = False
        line = 0
        for c in num_cols:
            col = control[c]
            #     print("{}, {}, {}, {}".format(c, depth, line, right))
            ax = plt.subplot2grid((depth, 2), (line, int(right)))
            g = col.dropna().plot.kde(ax=ax, title=str.title(c))
            g.get_xaxis().tick_bottom()
            g.get_yaxis().tick_left()
            if right:
                line += 1
            right = not right
        plt.tight_layout()
        plt.show()
        plt.clf()

    @staticmethod
    def show_categories(canonical: pa.Table, headers: [str, list]=None, d_types: [str, list]=None,
                         regex: [str, list]=None, drop: bool=None, capped_at: int=None, top=None):
        """"""
        cap = capped_at if isinstance(capped_at, int) else 1_000_000
        if canonical.num_rows*canonical.num_columns > cap > 0:
            sample = random.sample(range(canonical.num_rows), k=int(cap/canonical.num_columns))
            canonical = canonical.take(sample)
        canonical = Commons.filter_columns(canonical, headers=headers, d_types=d_types, regex=regex, drop=drop)
        cat_cols = Commons.filter_headers(canonical, d_types=[pa.string()])
        control = canonical.to_pandas()
        sns.set(style='darkgrid', color_codes=True)
        if len(cat_cols) == 1:
            c = cat_cols[0]
            width = control[c].nunique() + 1
            if width > 16:
                width = 16
            _ = plt.subplots(1, 1, figsize=(width, 4))
            _ = sns.countplot(c)
            _ = sns.countplot(x=c, data=control, palette="summer")
            _ = plt.xticks(rotation=-90)
            _ = plt.xlabel(str.title(c))
            _ = plt.ylabel('Count')
            title = "{} Categories".format(str.title(c))
            _ = plt.title(title, fontsize=16)
        else:
            wide_col, thin_col = [], []
            for c in cat_cols:
                if control[c].nunique() > 10:
                    wide_col += [c]
                else:
                    thin_col += [c]
            depth = len(wide_col) + int(round(len(thin_col) / 2, 0))
            _ = plt.figure(figsize=(20, 5 * depth))
            sns.set(style='darkgrid', color_codes=True)
            for c, i in zip(wide_col, range(len(wide_col))):
                ax = plt.subplot2grid((depth, 2), (i, 0), colspan=2)
                order = list(control[c].value_counts().index.values)
                if isinstance(top, int):
                    order = order[:top]
                _ = sns.countplot(x=c, data=control, ax=ax, order=order)
                _ = plt.xticks(rotation=-90)
                _ = plt.xlabel(str.title(c))
                _ = plt.ylabel('Count')
                title = "{} Categories".format(str.title(c))
                _ = plt.title(title, fontsize=16)
            right = False
            line = len(wide_col)
            for c in thin_col:
                ax = plt.subplot2grid((depth, 2), (line, int(right)))
                order = list(control[c].value_counts().index.values)
                _ = sns.countplot(x=c, data=control, ax=ax, order=order)
                _ = plt.xticks(rotation=-90)
                _ = plt.xlabel(str.title(c))
                _ = plt.ylabel('Count')
                title = "{} Categories".format(str.title(c))
                _ = plt.title(title, fontsize=16)
                if right:
                    line += 1
                right = not right

        _ = plt.tight_layout()
        _ = plt.show()
        _ = plt.clf()



import inspect
import pickle
from typing import Any
import numpy as np
import pandas as pd
import pyarrow as pa
from ds_capability.intent.common_intent import CommonsIntentModel
from ds_core.handlers.abstract_handlers import HandlerFactory
from ds_capability.intent.abstract_automl_intent import AbstractAutoMLIntentModel

__author__ = 'Darryl Oatridge'


class AutoMLIntent(AbstractAutoMLIntentModel, CommonsIntentModel):

    TRAIN_INTENT_LEVEL = 'train_level'
    PREDICT_INTENT_LEVEL = 'predict_level'

    def label_predict(self, canonical: pa.Table, model_name: str, *, id_header: str=None, save_intent: bool=None,
                      intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                      remove_duplicates: bool=None):
        """ Retrieves a trained model and applies it to the canonical, returning the canonical with prediction labels.
        This assumes a trained model with a predict function. if an ``id_header`` name is given, that column will be
        removed from the feature and reapplied with the predictions.

        :param canonical: the model canonical
        :param model_name: a unique name for the model
        :param id_header: (optional) the name of a header that is not a feature that uniquely identifies each row
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the level name that groups intent by a reference name
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: the canonical with a prediction.
        """
        # resolve intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # Code block for intent
        if self._pm.has_connector(model_name):
            canonical = self._get_canonical(canonical)
            handler = self._pm.get_connector_handler(model_name)
            model = handler.load_canonical()
            model = model.column(model_name).combine_chunks()
            model = pickle.loads(model[0].as_py())
            _id = None
            if isinstance(id_header, str) and id_header in canonical.column_names:
                _id = canonical.column(id_header).combine_chunks()
                canonical = canonical.drop_columns(id_header)
            features = np.asarray(canonical)
            score = model.predict(features).ravel()
            if isinstance(_id, pa.Array):
                return pa.table([_id, score], names=[id_header, 'predict'])
            return pa.table([score], names=['predict'])
        raise FileNotFoundError("The trained model cannot be found. Ensure the trained model has been added.")




import unittest
import os
from datetime import datetime
from pathlib import Path
import shutil
from pprint import pprint
import pyarrow as pa
import pyarrow.compute as pc
import pandas as pd
from ds_capability.components.commons import Commons

from ds_capability.intent.feature_engineer_intent import FeatureEngineerIntent

from ds_capability import FeatureEngineer, FeatureEngineer
from ds_core.properties.property_manager import PropertyManager

from ds_capability.components.discovery import DataDiscovery

# Pandas setup
pd.set_option('max_colwidth', 320)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 99)
pd.set_option('expand_frame_repr', True)


class DiscoveryTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        # clean out any old environments
        for key in os.environ.keys():
            if key.startswith('HADRON'):
                del os.environ[key]
        # Local Domain Contract
        os.environ['HADRON_PM_PATH'] = os.path.join('working', 'contracts')
        os.environ['HADRON_PM_TYPE'] = 'json'
        # Local Connectivity
        os.environ['HADRON_DEFAULT_PATH'] = Path('working/data').as_posix()
        # Specialist Component
        try:
            os.makedirs(os.environ['HADRON_PM_PATH'])
        except OSError:
            pass
        try:
            os.makedirs(os.environ['HADRON_DEFAULT_PATH'])
        except OSError:
            pass
        try:
            shutil.copytree('../_test_data', os.path.join(os.environ['PWD'], 'working/source'))
        except OSError:
            pass
        PropertyManager._remove_all()

    def tearDown(self):
        try:
            shutil.rmtree('working')
        except OSError:
            pass

    def test_interquartile_outliers(self):
        sb = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = sb.tools
        arr = pa.array([1,2,1,34,1,2,3])
        result = DataDiscovery.outliers_iqr(arr)
        self.assertEqual(([],[34]), result)

    def test_empirical_outliers(self):
        sb = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = sb.tools
        arr = pa.array([1,2,1,1,2,3]*100 +[34])
        result = DataDiscovery.outliers_empirical(arr)
        self.assertEqual(([],[34]), result)

    def test_data_dictionary(self):
        sb = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = sb.tools
        tbl = tools.get_synthetic_data_types(1_000)
        result = DataDiscovery.data_dictionary(tbl, stylise=False)
        self.assertEqual(['Attributes', 'DataType', 'Nulls', 'Dominate', 'Valid', 'Unique', 'Observations'], result.column_names)
        result = DataDiscovery.data_dictionary(tbl, stylise=False, ordered=True)
        pprint(result.column('Attributes'))

    def test_data_quality(self):
        sb = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = sb.tools
        tbl = tools.get_synthetic_data_types(100_000, extend=True)
        result = DataDiscovery.data_quality(tbl, stylise=True)
        pprint(result.to_string())

    def test_data_quality_ref(self):
        sb = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = sb.tools
        tbl = tools.get_synthetic_data_types(100_000, extend=True)
        result = DataDiscovery.data_quality(tbl, stylise=False)
        pprint(result.schema)

    def test_data_schema(self):
        sb = FeatureEngineer.from_memory()
        tools: FeatureEngineerIntent = sb.tools
        tbl = tools.get_synthetic_data_types(1_000, extend=True)
        result = DataDiscovery.data_schema(tbl, stylise=True)
        pprint(result.to_string())

    def test_raise(self):
        startTime = datetime.now()
        with self.assertRaises(KeyError) as context:
            env = os.environ['NoEnvValueTest']
        self.assertTrue("'NoEnvValueTest'" in str(context.exception))
        print(f"Duration - {str(datetime.now() - startTime)}")


if __name__ == '__main__':
    unittest.main()



from abc import abstractmethod
import pandas as pd
import pyarrow as pa
from ds_core.components.abstract_component import AbstractComponent
from ds_capability.components.commons import Commons
from ds_capability.components.discovery import DataDiscovery


# noinspection PyArgumentList
class AbstractCommonComponent(AbstractComponent):

    """
    An abstract common component class that contains the methods shared across all
    capabilities. This allows all capability instances to share common behavior in
    initialization, connectivity management, reporting and running the component.
    """

    @classmethod
    @abstractmethod
    def from_uri(cls, task_name: str, uri_pm_path: str, creator: str, uri_pm_repo: str=None, pm_file_type: str=None,
                 pm_module: str=None, pm_handler: str=None, pm_kwargs: dict=None, default_save=None,
                 reset_templates: bool=None, template_path: str=None, template_module: str=None,
                 template_source_handler: str=None, template_persist_handler: str=None, align_connectors: bool=None,
                 default_save_intent: bool=None, default_intent_level: bool=None, order_next_available: bool=None,
                 default_replace_intent: bool=None, has_contract: bool=None):
        return cls

    def run_component_pipeline(self, intent_levels: [str, int, list]=None, run_book: str=None, seed: int=None,
                               reset_changed: bool=None, has_changed: bool=None, **kwargs):
        """runs the synthetic component pipeline. By passing an int value as the canonical will generate a synthetic
        file of that size

        :param intent_levels: (optional) a single or list of intent levels to run
        :param run_book: (optional) a saved runbook to run
        :param seed: (optional) a seed value for this run
        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param kwargs: any additional kwargs
        """
        run_book = run_book if isinstance(run_book, str) and self.pm.has_run_book(run_book) else self.pm.PRIMARY_RUN_BOOK
        if isinstance(intent_levels, (str, int, list)):
            intent_levels = Commons.list_formatter(intent_levels)
        elif isinstance(run_book, str) and self.pm.has_run_book(book_name=run_book):
            intent_levels = self.pm.get_run_book(book_name=run_book)
        else:
            intent_levels = list(self.pm.get_intent().keys())
        canonical = None
        if self.pm.has_connector(self.CONNECTOR_SOURCE):
            canonical = self.load_source_canonical(reset_changed=reset_changed, has_changed=has_changed)
        for level in intent_levels:
            canonical = self.intent_model.run_intent_pipeline(canonical=canonical, intent_level=level, seed=seed, **kwargs)
        self.save_persist_canonical(canonical)
        return

    def load_source_canonical(self, reset_changed: bool=None, has_changed: bool=None, return_empty: bool=None,
                              **kwargs) -> pa.Table:
        """returns the contracted source data as a DataFrame

        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param return_empty: (optional) if has_changed is set, returns an empty canonical if set to True
        :param kwargs: arguments to be passed to the handler on load
        """
        return self.load_canonical(self.CONNECTOR_SOURCE, reset_changed=reset_changed, has_changed=has_changed,
                                   return_empty=return_empty, **kwargs)

    def load_canonical(self, connector_name: str, reset_changed: bool=None, has_changed: bool=None,
                       return_empty: bool=None, **kwargs) -> pa.Table:
        """returns the canonical of the referenced connector

        :param connector_name: the name or label to identify and reference the connector
        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param return_empty: (optional) if has_changed is set, returns an empty canonical if set to True
        :param kwargs: arguments to be passed to the handler on load
        """
        canonical = super().load_canonical(connector_name=connector_name, reset_changed=reset_changed,
                                           has_changed=has_changed, return_empty=return_empty, **kwargs)
        return canonical

    def load_persist_canonical(self, reset_changed: bool=None, has_changed: bool=None, return_empty: bool=None,
                               **kwargs) -> pa.Table:
        """loads the clean pandas.DataFrame from the clean folder for this contract

        :param reset_changed: (optional) resets the has_changed boolean to True
        :param has_changed: (optional) tests if the underline canonical has changed since last load else error returned
        :param return_empty: (optional) if has_changed is set, returns an empty canonical if set to True
        :param kwargs: arguments to be passed to the handler on load
        """
        return self.load_canonical(self.CONNECTOR_PERSIST, reset_changed=reset_changed, has_changed=has_changed,
                                   return_empty=return_empty, **kwargs)

    def save_persist_canonical(self, canonical, auto_connectors: bool=None, **kwargs):
        """Saves the canonical to the clean files folder, auto creating the connector from template if not set"""
        if auto_connectors if isinstance(auto_connectors, bool) else True:
            if not self.pm.has_connector(self.CONNECTOR_PERSIST):
                self.set_persist()
        self.persist_canonical(connector_name=self.CONNECTOR_PERSIST, canonical=canonical, **kwargs)

    def add_column_description(self, column_name: str, description: str, save: bool=None):
        """ adds a description note that is included in with the 'report_column_catalog'"""
        if isinstance(description, str) and description:
            self.pm.set_intent_description(level=column_name, text=description)
            self.pm_persist(save)
        return

    @staticmethod
    def quality_report(canonical: pa.Table, nulls_threshold: float=None, dom_threshold: float=None,
                     cat_threshold: int=None, stylise: bool=None):
        """ Analyses a dataset, passed as a DataFrame and returns a quality summary

        :param canonical: The table to view.
        :param cat_threshold: (optional) The threshold for the max number of unique categories. Default is 60
        :param dom_threshold: (optional) The threshold limit of a dominant value. Default 0.98
        :param nulls_threshold: (optional) The threshold limit of a nulls value. Default 0.9
        :param stylise: (optional) if the output is stylised
        """
        stylise = stylise if isinstance(stylise, bool) else True
        return DataDiscovery.data_quality(canonical=canonical, nulls_threshold=nulls_threshold,
                                          dom_threshold=dom_threshold, cat_threshold=cat_threshold, stylise=stylise)
    @staticmethod
    def canonical_report(canonical: pa.Table, headers: [str,list]=None, regex:[str,list]=None, d_types:list=None,
                         drop: bool=None, stylise: bool=None, display_width: int=None, ordered: bool=None):
        """The Canonical Report is a data dictionary of the canonical providing a reference view of the dataset's
        attribute properties

        :param canonical: the table to view
        :param headers: (optional) specific headers to display
        :param regex: (optional) specify header regex to display. regex matching is done using the Google RE2 library.
        :param d_types: (optional) a list of pyarrow DataType e.g [pa.string(), pa.bool_()]
        :param drop: (optional) if the headers are to be dropped and the remaining to display
        :param stylise: (optional) if True present the report stylised.
        :param display_width: (optional) the width of the observational display
        :param ordered: (optional) if the result should be in header order
        """
        stylise = stylise if isinstance(stylise, bool) else True
        tbl = Commons.filter_columns(canonical, headers=headers, regex=regex, d_types=d_types, drop=drop)
        return DataDiscovery.data_dictionary(canonical=tbl, stylise=stylise, display_width=display_width,
                                             ordered=ordered)

    @staticmethod
    def schema_report(canonical: pa.Table, headers: [str,list]=None, regex:[str,list]=None, d_types:list=None,
                      drop: bool=None, stylise: bool=True, table_cast: bool=None):
        """ presents the current canonical schema

        :param canonical: the table to view
        :param headers: (optional) specific headers to display
        :param regex: (optional) specify header regex to display. regex matching is done using the Google RE2 library.
        :param d_types: (optional) a list of pyarrow DataType e.g [pa.string(), pa.bool_()]
        :param drop: (optional) if the headers are to be dropped and the remaining to display
        :param stylise: (optional) if True present the report stylised.
        :param table_cast: (optional) if the column should try to be cast to its type
        """
        stylise = stylise if isinstance(stylise, bool) else True
        table_cast = table_cast if isinstance(table_cast,bool) else True
        tbl = Commons.filter_columns(canonical, headers=headers, regex=regex, d_types=d_types, drop=drop)
        return DataDiscovery.data_schema(canonical=tbl, table_cast=table_cast, stylise=stylise)

    @staticmethod
    def table_report(canonical: pa.Table, head: int=None):
        """ Creates a report from a pyarrow table in a tabular form

        :param canonical: the table to view
        :param head: The number of rows to show.
        """
        return Commons.table_report(canonical, head=head)

    def report_task(self, stylise: bool=True):
        """ generates a report on the source contract

        :param stylise: (optional) returns a stylised DataFrame with formatting
        :return: pa.Table
        """
        report = self.pm.report_task_meta()
        df = pd.DataFrame.from_dict(data=report, orient='index').reset_index()
        df.columns = ['name', 'value']
        # sort out any values that start with a $ as it throws formatting
        for c in df.columns:
            df[c] = [f"{x[1:]}" if str(x).startswith('$') else x for x in df[c]]
        if stylise:
            return Commons.report(df, index_header='name')
        return pa.Table.from_pandas(df)

    def report_connectors(self, connector_filter: [str, list]=None, inc_pm: bool=None, inc_template: bool=None,
                          stylise: bool=True):
        """ generates a report on the source contract

        :param connector_filter: (optional) filters on the connector name.
        :param inc_pm: (optional) include the property manager connector
        :param inc_template: (optional) include the template connectors
        :param stylise: (optional) returns a stylised DataFrame with formatting
        :return: pa.Table
        """
        report = self.pm.report_connectors(connector_filter=connector_filter, inc_pm=inc_pm,
                                           inc_template=inc_template)
        df = pd.DataFrame.from_dict(data=report)
        # sort out any values that start with a $ as it throws formatting
        for c in df.columns:
            df[c] = [f"{x[1:]}" if str(x).startswith('$') else x for x in df[c]]
        if stylise:
            return Commons.report(df, index_header='connector_name')
        return pa.Table.from_pandas(df)

    def report_intent_description(self, action: [str, list]=None, stylise: bool=True):
        """ generates a report on the intent action notes

        :param action: (optional) filters on specific action names.
        :param stylise: (optional) returns a stylised DataFrame with formatting
        :return: pa.Table
        """
        stylise = True if not isinstance(stylise, bool) else stylise
        df = pd.DataFrame.from_dict(data=self.pm.report_intent(levels=action, as_description=True,
                                                               level_label='column_name'))
        if stylise:
            return Commons.report(df, index_header='column_name')
        return pa.Table.from_pandas(df)

    def report_run_book(self, stylise: bool=True):
        """ generates a report on all the intent

        :param stylise: returns a stylised dataframe with formatting
        :return: pa.Table
        """
        df = pd.DataFrame.from_dict(data=self.pm.report_run_book())
        if stylise:
            return Commons.report(df, index_header='name')
        return pa.Table.from_pandas(df)

    def report_environ(self, stylise: bool=True):
        """ generates a report on all the intent

        :param stylise: returns a stylised dataframe with formatting
        :return: pa.Table
        """
        df = pd.DataFrame.from_dict(data=super().report_environ(), orient='index').reset_index()
        df.columns = ["environ", "value"]
        if stylise:
            return Commons.report(df, bold='environ')
        return pa.Table.from_pandas(df)

    def report_intent(self, levels: [str, int, list]=None, stylise: bool=True):
        """ generates a report on all the intent

        :param levels: (optional) a filter on the levels. passing a single value will report a single parameterised view
        :param stylise: (optional) returns a stylised dataframe with formatting
        :return: pa.Table
        """
        if isinstance(levels, (int, str)):
            df = pd.DataFrame.from_dict(data=self.pm.report_intent_params(level=levels))
            if stylise:
                return Commons.report(df, index_header='order')
        df = pd.DataFrame.from_dict(data=self.pm.report_intent(levels=levels))
        if stylise:
            return Commons.report(df, index_header='level')
        return pa.Table.from_pandas(df)

    def report_notes(self, catalog: [str, list]=None, labels: [str, list]=None, regex: [str, list]=None,
                     re_ignore_case: bool=False, stylise: bool=True, drop_dates: bool=False):
        """ generates a report on the notes

        :param catalog: (optional) the catalog to filter on
        :param labels: (optional) s label or list of labels to filter on
        :param regex: (optional) a regular expression on the notes
        :param re_ignore_case: (optional) if the regular expression should be case-sensitive
        :param stylise: (optional) returns a stylised dataframe with formatting
        :param drop_dates: (optional) excludes the 'date' column from the report
        :return: pa.Table
        """
        report = self.pm.report_notes(catalog=catalog, labels=labels, regex=regex, re_ignore_case=re_ignore_case,
                                      drop_dates=drop_dates)
        df = pd.DataFrame.from_dict(data=report)
        if stylise:
            return Commons.report(df, index_header='section', bold='label')
        return pa.Table.from_pandas(df)

    @classmethod
    def __dir__(cls):
        """returns the list of available methods associated with the parameterized intent"""
        rtn_list = []
        for m in dir(cls):
            if not m.startswith('_'):
                rtn_list.append(m)
        return rtn_list



import inspect
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
from ds_capability.components.discovery import DataDiscovery
from scipy.stats import stats

from ds_capability.components.commons import Commons
from ds_capability.intent.common_intent import CommonsIntentModel
from ds_capability.intent.abstract_feature_transform_intent import AbstractFeatureTransformIntentModel


class FeatureTransformIntent(AbstractFeatureTransformIntentModel, CommonsIntentModel):
    """This class represents feature transformation intent actions whereby features are converted
    from one format or structure to another. This includes, scaling, encoding,
    discretization and activation trigger algorithms.
    """

    def activate_sigmoid(self, canonical: pa.Table, header: str, precision: int=None, seed: int=None,
                         save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                         replace_intent: bool=None, remove_duplicates: bool=None):
        """Activation functions play a crucial role in the backpropagation algorithm, which is the primary
        algorithm used for training neural networks. During backpropagation, the error of the output is
        propagated backwards through the network, and the weights of the network are updated based on this
        error. The activation function is used to introduce non-linearity into the output of a neural network
        layer.

        Logistic Sigmoid a.k.a logit, tmaps any input value to a value between 0 and 1, making it useful for
        binary classification problems and is defined as f(x) = 1/(1+exp(-x))

        The sigmoid function has an S-shaped curve, and it asymptotically approaches 0 as z approaches
        negative infinity and approaches 1 as z approaches positive infinity. This property makes it
        useful for binary classification problems, where the goal is to classify inputs into one of
        two categories.

        :param canonical: a pa.Table as the reference dataframe
        :param header: the header in the Table to correlate
        :param precision: (optional) how many decimal places. default to 3
        :param seed: (optional) the random seed. defaults to current datetime
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the intent level that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intend code block on the canonical
        canonical = self._get_canonical(canonical)
        if not isinstance(header, str) or header not in canonical.column_names:
            raise ValueError(f"The header '{header}' can't be found in the canonical table")
        if canonical.column(header).null_count > 0:
            raise ValueError(f"The values in '{header}' can not contain nulls")
        _seed = seed if isinstance(seed, int) else self._seed()
        precision = precision if isinstance(precision, int) else 5
        npv = canonical.column(header).to_numpy()
        arr = pa.array(np.round(1 / (1 + np.exp(-npv)), precision))
        return Commons.table_append(canonical, pa.table([arr], names=[header]))

    def activate_tanh(self, canonical: pa.Table, header: str, precision: int=None, seed: int=None,
                      save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                      replace_intent: bool=None, remove_duplicates: bool=None):
        """Activation functions play a crucial role in the backpropagation algorithm, which is the primary
        algorithm used for training neural networks. During backpropagation, the error of the output is
        propagated backwards through the network, and the weights of the network are updated based on this
        error. The activation function is used to introduce non-linearity into the output of a neural network
        layer.

        Tangent Hyperbolic (tanh) function is a shifted and stretched version of the Sigmoid function but maps
        the input values to a range between -1 and 1. and is defined as f(x) = (exp(x)-exp(-x))/(exp(x)+exp(-x))

        Similar to the logistic sigmoid function, the hyperbolic tangent function has an S-shaped curve,
        but it maps its input to values between -1 and 1. This function is advantageous because it has
        desirable properties, such as being zero-centered (meaning that the average of its output is
        close to zero), which can help with the convergence of optimization algorithms during training.

        :param canonical: a pa.Table as the reference dataframe
        :param header: the header in the Table to correlate
        :param precision: (optional) how many decimal places. default to 3
        :param seed: (optional) the random seed. defaults to current datetime
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the intent level that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intend code block on the canonical
        canonical = self._get_canonical(canonical)
        if not isinstance(header, str) or header not in canonical.column_names:
            raise ValueError(f"The header '{header}' can't be found in the canonical table")
        if canonical.column(header).null_count > 0:
            raise ValueError(f"The values in '{header}' can not contain nulls")
        _seed = seed if isinstance(seed, int) else self._seed()
        precision = precision if isinstance(precision, int) else 5
        npv = canonical.column(header).to_numpy()
        arr = pa.array(np.round((np.exp(npv) - np.exp(-npv)) / (np.exp(npv) + np.exp(-npv)), precision))
        return Commons.table_append(canonical, pa.table([arr], names=[header]))

    def activate_relu(self, canonical: pa.Table, header: str, precision: int=None, seed: int=None,
                      save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                      replace_intent: bool=None, remove_duplicates: bool=None):
        """Activation functions play a crucial role in the backpropagation algorithm, which is the primary
        algorithm used for training neural networks. During backpropagation, the error of the output is
        propagated backwards through the network, and the weights of the network are updated based on this
        error. The activation function is used to introduce non-linearity into the output of a neural network
        layer.

        Rectified Linear Unit (ReLU) function. is the most popular activation function, which replaces negative
        values with zero and keeps the positive values unchanged. and is defined as f(x) = x * (x > 0)

        :param canonical: a pa.Table as the reference dataframe
        :param header: the header in the Table to correlate
        :param precision: (optional) how many decimal places. default to 3
        :param seed: (optional) the random seed. defaults to current datetime
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the intent level that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intend code block on the canonical
        canonical = self._get_canonical(canonical)
        if not isinstance(header, str) or header not in canonical.column_names:
            raise ValueError(f"The header '{header}' can't be found in the canonical table")
        if canonical.column(header).null_count > 0:
            raise ValueError(f"The values in '{header}' can not contain nulls")
        _seed = seed if isinstance(seed, int) else self._seed()
        precision = precision if isinstance(precision, int) else 5
        npv = canonical.column(header).to_numpy()
        arr = pa.array(np.round(npv * (npv > 0), precision))
        return Commons.table_append(canonical, pa.table([arr], names=[header]))

    def encode_date_integer(self, canonical: pa.Table, headers: [str, list]=None, prefix=None, day_first: bool=None,
                            year_first: bool=None, seed: int=None, save_intent: bool=None,
                            intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                            remove_duplicates: bool=None):
        """ date encoding to integer replaces dates for integer values.

        :param canonical: pyarrow Table
        :param headers: the header(s) to apply encoding too
        :param prefix: (optional) a str to prefix the column
        :param day_first: (optional) if the dates given are day first format. Default to True
        :param year_first: (optional) if the dates given are year first. Default to False
        :param seed: seed: (optional) a seed value for the random function: default to None
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the intent level that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intend code block on the canonical
        canonical = self._get_canonical(canonical)
        prefix = prefix if isinstance(prefix, str) else ''
        headers = Commons.list_formatter(headers) if isinstance(headers,(str, list)) else canonical.column_names
        _ = self._seed() if seed is None else seed
        tbl = None
        for n in headers:
            c = canonical.column(n).combine_chunks()
            if not (pa.types.is_timestamp(c.type) or pa.types.is_time(c.type)):
                continue
            column = pa.array(Commons.date2value(c.to_pylist()), pa.int64())
            new_header = f"{prefix}{n}"
            tbl = Commons.table_append(tbl, pa.table([column], names=[new_header]))
        if not tbl:
            return canonical
        return Commons.table_append(canonical, tbl)

    def encode_category_integer(self, canonical: pa.Table, headers: [str, list]=None, ordinal: bool=None,
                                label_count: int=None, prefix=None, seed: int=None, save_intent: bool=None,
                                intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                                remove_duplicates: bool=None):
        """ Integer encoding replaces the categories by digits from 1 to n, where n is the number of distinct
        categories of the variable. Integer encoding can be either nominal or ordinal.

        Nominal data is categorical variables without any particular order between categories. This means that
        the categories cannot be sorted and there is no natural order between them.

        Ordinal data represents categories with a natural, ordered relationship between each category. This means
        that the categories can be sorted in either ascending or descending order. In order to encode integers as
        ordinal, a ranking must be provided.

        If ranking is given, the return will be ordinal values based on the ranking order of the list. If a
        categorical value is not found in the list it is grouped with other missing values and given the last
        ranking. This is known as rare-label encoding.

        :param canonical: pyarrow Table
        :param headers: the header(s) to apply encoding too
        :param ordinal: (optional) if the integer encoder is ordinal
        :param label_count: (optional) if the ordinal is rare-label, the number of categories to rank
        :param prefix: (optional) a str to prefix the column
        :param seed: seed: (optional) a seed value for the random function: default to None
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the intent level that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intend code block on the canonical
        canonical = self._get_canonical(canonical)
        prefix = prefix if isinstance(prefix, str) else ''
        headers = Commons.list_formatter(headers) if isinstance(headers, (str, list)) else canonical.column_names
        _ = self._seed() if seed is None else seed
        tbl = None
        for header in headers:
            column = canonical.column(header).combine_chunks()
            if isinstance(label_count, int): # rare-label
                if not pa.types.is_dictionary(column.type):
                    column = column.dictionary_encode()
                full_rank = column.dictionary.sort().to_pylist()
                rank = full_rank[:label_count]
                missing = full_rank[label_count:]
                values = list(range(len(rank)))
                values = values + ([len(rank)] * (len(full_rank) - len(rank)))
                mapper = dict(zip(full_rank, values))
                s_column = column.to_pandas()
                column =  pa.Array.from_pandas(s_column.replace(mapper))
            elif isinstance(ordinal, bool) and ordinal: # ordinal
                if not pa.types.is_dictionary(column.type):
                    column = column.dictionary_encode()
                rank = column.dictionary.sort().to_pylist()
                values = list(range(len(rank)))
                mapper = dict(zip(rank, values))
                s_column = column.to_pandas()
                column = pa.Array.from_pandas(s_column.replace(mapper))
            else: # nominal
                if not pa.types.is_dictionary(column.type):
                    column = column.dictionary_encode()
                column = pa.array(column.indices, pa.int64())
            if pa.types.is_dictionary(column.type):
                column = column.dictionary_decode()
            new_header = f"{prefix}{header}"
            tbl = Commons.table_append(tbl, pa.table([column], names=[new_header]))
        if not tbl:
            return canonical
        return Commons.table_append(canonical, tbl)

    def encode_category_one_hot(self, canonical: pa.Table, headers: [str, list]=None, prefix=None, data_type: pa.Table=None,
                                prefix_sep: str=None, dummy_na: bool = False, drop_first: bool = False, seed: int=None,
                                save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                                replace_intent: bool=None, remove_duplicates: bool=None) -> pa.Table:
        """ encodes categorical data types, One hot encoding, consists in encoding each categorical variable with
        different boolean variables (also called dummy variables) which take values 0 or 1, indicating if a category
        is present in an observation.

        :param canonical: pyarrow Table
        :param headers: the header(s) to apply encoding too
        :param prefix: str, list of str, or dict of str, String to append Table intent levels, with equal length.
        :param prefix_sep: str separator, default '_'
        :param dummy_na: Add a column to indicate null values, if False nullss are ignored.
        :param drop_first:  Whether to get k-1 dummies out of k categorical levels by removing the first level.
        :param data_type: Data type for new columns. Only a single dtype is allowed.
        :param seed: seed: (optional) a seed value for the random function: default to None
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the intent level that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intend code block on the canonical
        canonical = self._get_canonical(canonical)
        headers = Commons.list_formatter(headers) if isinstance(headers, (str, list)) else canonical.column_names
        _ = self._seed() if seed is None else seed
        prefix_sep = prefix_sep if isinstance(prefix_sep, str) else "_"
        dummy_na = dummy_na if isinstance(dummy_na, bool) else False
        drop_first = drop_first if isinstance(drop_first, bool) else False
        d_type = data_type if data_type else pa.int64()
        dummies = pd.get_dummies(canonical.to_pandas(), columns=headers, prefix=prefix, prefix_sep=prefix_sep,
                              dummy_na=dummy_na, drop_first=drop_first, dtype=data_type)
        return pa.Table.from_pandas(dummies)

    def scale_normalize(self, canonical: pa.Table, headers: [str, list]=None, scalar: [tuple, str]=None,
                        prefix: str=None, precision: int=None, seed: int=None, save_intent: bool=None,
                        intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                        remove_duplicates: bool=None):
        """ Normalization of continuous data using either Min-Max Scaling or Robust Scaling, dependent on the
        scalar passed.

        Min-Max Scaling: scales the data such that it falls within a specified range, typically (0,1) as default if
        no value is passed, or (min, max) tuple. This method is suitable when your data should be uniformly
        distributed across the specified range.

        Robust Scaling: Robust scaling is similar to min-max scaling but is less sensitive to outliers. It uses
        the interquartile range (0.25, 0.75) to scale the data. This is used if the scalar equals 'robust'.
        Robust scaling is particularly useful when your dataset contains outliers or when the underlying data
        distribution is not necessarily Gaussian.

        :param canonical: pyarrow Table
        :param headers: the header(s) to apply scaling too
        :param scalar: (optional) a tuple scalar representing min and max values or 'robust' for interquartile scaling
        :param prefix: (optional) a str prefix for generated headers
        :param precision: (optional) how many decimal places. default to 3
        :param seed: (optional) a seed value for the random function: default to None
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the intent level that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intend code block on the canonical
        canonical = self._get_canonical(canonical)
        headers = Commons.list_formatter(headers) if isinstance(headers, (str, list)) else canonical.column_names
        _seed = seed if isinstance(seed, int) else self._seed()
        scalar = scalar if isinstance(scalar, (tuple, str)) else (0, 1)
        tbl = None
        for n in headers:
            c = canonical.column(n).combine_chunks()
            if not (pa.types.is_floating(c.type) or pa.types.is_integer(c.type)):
                continue
            precision = precision if isinstance(precision, int) else Commons.column_precision(c)+2
            s_values = c.to_pandas()
            null_idx = s_values[s_values.isna()].index
            s_values = s_values.fillna(0)
            if isinstance(scalar, str) and scalar == 'robust':
                if s_values.max() == s_values.min():
                    s_values = pd.Series([0.5] * canonical.num_rows)
                else:
                    s_values = pd.Series(Commons.list_normalize_robust(s_values.to_list(), precision))
            elif s_values.max() == s_values.min():
                s_values = pd.Series([np.mean([scalar[0], scalar[1]])] * canonical.num_rows)
            else:
                if scalar[0] >= scalar[1] or len(scalar) != 2:
                    scalar = (0,1)
                s_values = pd.Series(Commons.list_normalize(s_values.to_list(), scalar[0], scalar[1]))
            s_values = s_values.round(precision)
            if null_idx.size > 0:
                s_values.iloc[null_idx] = np.nan
            new_header = f"{prefix}{n}"
            tbl = Commons.table_append(tbl, pa.table([s_values], names=[n]))
        if not tbl:
            return canonical
        return Commons.table_append(canonical, tbl)

    def scale_standardize(self, canonical: pa.Table, headers: [str, list]=None, prefix: str=None, precision: int=None,
                          seed: int=None, save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                          replace_intent: bool=None, remove_duplicates: bool=None):
        """ Z-Score Standardization (Standard Scaling). This method transforms the data to have a mean of 0
        and a standard deviation of 1. It's particularly useful when your data follows a Gaussian (normal)
        distribution. This transformation makes it easier to compare and work with features that may have
        different scales and ensures that they contribute equally to model training.

        :param canonical: pyarrow Table
        :param headers: the header(s) to apply scaling too
        :param prefix: (optional) a str prefix for generated headers
        :param precision: (optional) how many decimal places. default to 3
        :param seed: seed: (optional) a seed value for the random function: default to None
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the intent level that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intend code block on the canonical
        canonical = self._get_canonical(canonical)
        headers = Commons.list_formatter(headers) if isinstance(headers, (str, list)) else canonical.column_names
        _seed = seed if isinstance(seed, int) else self._seed()
        tbl = None
        for n in headers:
            c = canonical.column(n).combine_chunks()
            if not (pa.types.is_floating(c.type) or pa.types.is_integer(c.type)):
                continue
            precision = precision if isinstance(precision, int) else Commons.column_precision(c)+2
            s_values = c.to_pandas()
            null_idx = s_values[s_values.isna()].index
            s_values = s_values.fillna(0)
            s_values = pd.Series(Commons.list_standardize(s_values.to_list()))
            s_values = s_values.round(precision)
            if null_idx.size > 0:
                s_values.iloc[null_idx] = np.nan
            new_header = f"{prefix}{n}"
            tbl = Commons.table_append(tbl, pa.table([s_values], names=[n]))
        if not tbl:
            return canonical
        return Commons.table_append(canonical, tbl)

    def scale_transform(self, canonical: pa.Table, transform: str, headers: [str, list]=None, prefix: str=None,
                        precision: int=None, seed: int=None, save_intent: bool=None, intent_level: [int, str]=None,
                        intent_order: int=None, replace_intent: bool=None, remove_duplicates: bool=None):
        """ Log, sqrt (square root), cbrt (cube root), Box-Cox, and Yeo-Johnson transformations. These are techniques
        used to modify the distribution or scale of continuous data to make the data conform more closely to a
        normal distribution or to stabilize the variance.

        Log Transformation (log): The log transformation involves taking the natural logarithm of each data point.
        It is particularly useful when dealing with data that follows an exponential distribution or has a
        right-skewed distribution. The log transformation can help make the data more symmetric and stabilize
        the variance.

        Square Root Transformation (sqrt): The square root transformation involves taking the square root of
        each data point. It is often used when dealing with count data or data with a right-skewed distribution.
        Like the log transformation, the square root transformation can make the data more symmetric and stabilize
        the variance.

        Cube Root Transformation (cbrt): The cube root transformation involves taking the cube root of each
        data point. It is another option for stabilizing variance and making data less skewed, especially when
        dealing with positive data with right-skewed distributions.

        Box-Cox Transformation: The Box-Cox transformation is a family of power transformations that includes
        both the log and square root transformations as special cases. Can only work with positive values

        Yeo-Johnson Transformation: The Yeo-Johnson transformation is an extension of the Box-Cox transformation
        that allows for the transformation of data with both positive and negative values.

        These transformations are often applied to address issues like non-normality, heteroscedasticity (unequal variance),
        and skewed distributions in the data.

        :param canonical: pyarrow Table
        :param headers: the header(s) to apply scaling to
        :param transform: transform function, 'log', 'sqrt', 'cbrt', 'boxcox' or 'yeojohnson'
        :param prefix: (optional) a str prefix for generated headers
        :param precision: (optional) how many decimal places. default to 3
        :param seed: seed: (optional) a seed value for the random function: default to None
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the intent level that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intend code block on the canonical
        canonical = self._get_canonical(canonical)
        headers = Commons.list_formatter(headers) if isinstance(headers, (str, list)) else canonical.column_names
        _seed = seed if isinstance(seed, int) else self._seed()
        tbl = None
        for n in headers:
            c = canonical.column(n).combine_chunks()
            if not (pa.types.is_floating(c.type) or pa.types.is_integer(c.type)):
                continue
            precision = precision if isinstance(precision, int) else Commons.column_precision(c)+2
            s_values = c.to_pandas()
            null_idx = s_values[s_values.isna()].index
            s_values = s_values.fillna(0)
            if transform == 'log':
                s_values = np.log(s_values)
            elif transform == 'sqrt':
                s_values = np.sqrt(s_values)
            elif transform == 'cbrt':
                s_values = np.cbrt(s_values)
            elif transform == 'boxcox' or transform.startswith('box'):
                bc, _ = stats.boxcox(s_values.to_numpy())
                s_values = pd.Series(bc)
            elif transform == 'yeojohnson' or transform.startswith("yeo"):
                yj, _ = stats.yeojohnson(s_values.to_numpy())
                s_values = pd.Series(yj)
            else:
                raise ValueError(f"The transformer {transform} is not recognized. See contacts notes for reference")
            s_values = s_values.round(precision)
            if null_idx.size > 0:
                s_values.iloc[null_idx] = np.nan
            new_header = f"{prefix}{n}"
            tbl = Commons.table_append(tbl, pa.table([s_values], names=[n]))
        if not tbl:
            return canonical
        return Commons.table_append(canonical, tbl)

    def scale_mapping(self, canonical: pa.Table, numerator: str, denominator: str, prefix: str=None,
                      precision: int=None, to_header: str=None, seed: int=None, save_intent: bool=None,
                      intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                      remove_duplicates: bool=None):
        """ Scales a numeric column against another numeric column providing mapped standardisation.
        An example might be a churn problem for a bank with the hypothesis that older customers are
        more like to stay. Therefore, scaling 'Tenure' against 'age' gives us a mapped scalar of
        the hypothesis.

        :param canonical: pyarrow Table
        :param numerator: the column name of the numerator
        :param denominator:  the column name of the denominator
        :param prefix: (optional) a str prefix for generated headers
        :param precision: (optional) how many decimal places. default to 3
        :param to_header: (optional) an optional name to call the column
        :param seed: seed: (optional) a seed value for the random function: default to None
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the intent level that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: a pa.Table
        """
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # intend code block on the canonical
        canonical = self._get_canonical(canonical)
        _seed = seed if isinstance(seed, int) else self._seed()
        a = canonical.column(numerator).cast(pa.float64())
        b = canonical.column(denominator)
        precision = precision if isinstance(precision, int) else Commons.column_precision(a) + 3
        arr = pc.round(pc.divide(a, b), precision)
        arr = pc.if_else(pc.equal(arr, np.inf), 0, arr)
        to_header = to_header if isinstance(to_header, str) else numerator
        if to_header == numerator:
            canonical = canonical.drop_columns([numerator, denominator])
        return Commons.table_append(canonical, pa.table([arr], names=[to_header]))

    def discrete_intervals(self, canonical: pa.Table, header: str, interval: [int, list]=None, categories: list=None,
                           to_header: str=None, precision: int=None, duplicates: str=None, seed: int=None,
                           save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                           replace_intent: bool=None, remove_duplicates: bool=None):
        """ Converts continuous values into discrete values through interval categorisation based on
        quantile discretization. Intervals can be either a single value representing the number of
        quantiles or an ordered list of quantiles between 0 and 1, e.g. [0, .25, .5, .75, 1.]

        If intervals are not unique duplicates provide a strategy to 'raise' a ValueError, 'drop'
        duplicate bins or 'rank' values from 1 to n before binning. Ranking might be used for sparse
        date with predominant zeros as an example.

        :param canonical: a pa.Table as the reference table
        :param header: the header in the Table to correlate
        :param interval: (optional) the granularity of the analysis across the range. Default is 3
                int passed - represents the number of periods
                list[int] - specific interval periods e.g []

        :param to_header: (optional) an optional name to call the column
        :param precision: (optional) The precision of the range and boundary values. by default set to 5.
        :param categories: (optional) a set of labels the same length as the intervals to name the categories
        :param duplicates: (optional) If intervals are not unique, 'raise' ValueError, 'drop' intervals or 'rank' values
        :param seed: (optional) the random seed. defaults to current datetime
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the column name that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: an equal length list of correlated values
        """
        self._set_intend_signature( self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                    intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                    remove_duplicates=remove_duplicates, save_intent=save_intent)
        # remove intent params
        canonical = self._get_canonical(canonical)
        header = self._extract_value(header)
        to_header = self._extract_value(to_header)
        if not isinstance(header, str) or header not in canonical.column_names:
            raise ValueError(f"The header '{header}' can't be found in the canonical headers")
        interval = interval if isinstance(interval, (int, list)) else 5
        n = interval if isinstance(interval, int) else len(interval)
        categories = categories if isinstance(categories, list) else range(1, n+1) if isinstance(interval, int) else range(1, n)
        precision = precision if isinstance(precision, int) else 3
        duplicates = duplicates if isinstance(duplicates, str) and duplicates in ['drop', 'rank'] else 'raise'
        seed = seed if isinstance(seed, int) else self._seed()
        c = canonical.column(header).combine_chunks()
        if not (pa.types.is_floating(c.type) or pa.types.is_integer(c.type)):
            raise ValueError(f"The header '{header}' value type must be numerical, '{c.type}' was passed")
        s = c.to_pandas()
        if duplicates == 'rank':
            s = s.rank(method="first")
            duplicates = 'raise'
        s = pd.cut(s, interval, labels=categories, precision=precision, duplicates=duplicates)
        c = pa.Array.from_pandas(s)
        if pa.types.is_dictionary(c.type):
            c = c.dictionary_decode()
        to_header = to_header if isinstance(to_header, str) else header
        return Commons.table_append(canonical, pa.table([c], names=[to_header]))

    def discrete_quantiles(self, canonical: pa.Table, header: str, interval: [int, list]=None, categories: list=None,
                           to_header: str=None, precision: int=None, duplicates: str=None, seed: int=None,
                           save_intent: bool=None, intent_level: [int, str]=None, intent_order: int=None,
                           replace_intent: bool=None, remove_duplicates: bool=None):
        """ Converts continuous values into discrete values through interval categorisation based on
        interval bins. Intervals can be either a single value number of bins or a list of bin edges
        allowing for non-uniform width.

        If intervals are not unique duplicates provide a strategy to 'raise' a ValueError, 'drop'
        duplicate bins or 'rank' values from 1 to n before binning. Ranking might be used for sparse
        date with predominant zeros as an example.

        :param canonical: a pa.Table as the reference table
        :param header: the header in the Table to correlate
        :param interval: (optional) the granularity of the analysis across the range. Default is 3
                int passed - represents the number of periods
                list[float] - the percentile or quantities, All should fall between 0 and 1

        :param to_header: (optional) an optional name to call the column
        :param precision: (optional) The precision of the range and boundary values. by default set to 5.
        :param categories: (optional)  a set of labels the same length as the intervals to name the categories
        :param duplicates: (optional) If intervals are not unique, 'raise' ValueError, 'drop' intervals or 'rank' values
        :param seed: (optional) the random seed. defaults to current datetime
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the column name that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: an equal length list of correlated values
        """
        self._set_intend_signature( self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                    intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                    remove_duplicates=remove_duplicates, save_intent=save_intent)
        # remove intent params
        canonical = self._get_canonical(canonical)
        header = self._extract_value(header)
        to_header = self._extract_value(to_header)
        if not isinstance(header, str) or header not in canonical.column_names:
            raise ValueError(f"The header '{header}' can't be found in the canonical headers")
        interval = interval if isinstance(interval, (int, list)) else 4
        n = interval if isinstance(interval, int) else len(interval)
        categories = categories if isinstance(categories, list) else False
        precision = precision if isinstance(precision, int) else 3
        duplicates = duplicates if isinstance(duplicates, str) and duplicates in ['drop', 'rank'] else 'raise'
        seed = seed if isinstance(seed, int) else self._seed()
        c = canonical.column(header).combine_chunks()
        if not (pa.types.is_floating(c.type) or pa.types.is_integer(c.type)):
            raise ValueError(f"The header '{header}' value type must be numerical, '{c.type}' was passed")
        s = c.to_pandas()
        if duplicates == 'rank':
            s = s.rank(method="first")
            duplicates = 'raise'
        s = pd.qcut(s, interval, labels=categories, precision=precision, duplicates=duplicates)
        if categories is False:
            s += 1
        c = pa.Array.from_pandas(s)
        if pa.types.is_dictionary(c.type):
            c = c.dictionary_decode()
        to_header = to_header if isinstance(to_header, str) else header
        return Commons.table_append(canonical, pa.table([c], names=[to_header]))

    def discrete_custom(self, canonical: pa.Table, header: str, interval: [int, float, list]=None,
                        lower: [int, float]=None, upper: [int, float]=None, categories: list=None,
                        to_header: str=None, precision: int=None, seed: int=None, save_intent: bool=None,
                        intent_level: [int, str]=None, intent_order: int=None, replace_intent: bool=None,
                        remove_duplicates: bool=None) -> pa.Table:
        """ converts continuous representation into discrete representation through interval
        categorisation based on custom intervals

        :param canonical: a pa.Table as the reference table
        :param header: the header in the Table to correlate
        :param interval: (optional) the granularity of the analysis across the range. Default is 3
                float passed - the length of each interval
                list[tuple] - specific interval periods e.g []

        :param lower: (optional) the lower limit of the number value. Default min()
        :param upper: (optional) the upper limit of the number value. Default max()
        :param to_header: (optional) an optional name to call the column
        :param precision: (optional) The precision of the range and boundary values. by default set to 5.
        :param categories: (optional)  a set of labels the same length as the intervals to name the categories
        :param seed: (optional) the random seed. defaults to current datetime
        :param save_intent: (optional) if the intent contract should be saved to the property manager
        :param intent_level: (optional) the column name that groups intent to create a column
        :param intent_order: (optional) the order in which each intent should run.
                    - If None: default's to -1
                    - if -1: added to a level above any current instance of the intent section, level 0 if not found
                    - if int: added to the level specified, overwriting any that already exist

        :param replace_intent: (optional) if the intent method exists at the level, or default level
                    - True - replaces the current intent method with the new
                    - False - leaves it untouched, disregarding the new intent

        :param remove_duplicates: (optional) removes any duplicate intent in any level that is identical
        :return: an equal length list of correlated values
        """
        # intent persist options
        self._set_intend_signature(self._intent_builder(method=inspect.currentframe().f_code.co_name, params=locals()),
                                   intent_level=intent_level, intent_order=intent_order, replace_intent=replace_intent,
                                   remove_duplicates=remove_duplicates, save_intent=save_intent)
        # remove intent params
        canonical = self._get_canonical(canonical)
        header = self._extract_value(header)
        to_header  = self._extract_value(to_header)
        if not isinstance(header, str) or header not in canonical.column_names:
            raise ValueError(f"The header '{header}' can't be found in the canonical headers")
        seed = seed if isinstance(seed, int) else self._seed()
        rtn_arr = DataDiscovery.to_discrete_intervals(column=canonical.column(header), granularity=interval,
                                                      lower=lower, upper=upper, categories=categories,
                                                      precision=precision)
        to_header = to_header if isinstance(to_header, str) else header
        return Commons.table_append(canonical, pa.table([rtn_arr.dictionary_encode()], names=[to_header]))




# bring definitions to the top level
from ds_capability.components.feature_select import FeatureSelect
from ds_capability.components.feature_engineer import FeatureEngineer
from ds_capability.components.feature_transform import FeatureTransform
from ds_capability.components.feature_build import FeatureBuild
from ds_capability.components.automl import AutoML
from ds_capability.components.controller import Controller


# release version number picked up in the setup.py
__version__ = '0.20.5'
